# > \brief \b SROTG
#
#  =========== DOCUMENTATION ===========
#
# Online html documentation available at
#            http://www.netlib.org/lapack/explore-html/
#
#  Definition:
#  ===========
#
#       def SROTG(SA,SB,C,S)
#
#       .. Scalar Arguments ..
#       REAL C,S,SA,SB
#       ..
#
#
# > \par Purpose:
#  =============
# >
# > \verbatim
# >
# >    SROTG construct givens plane rotation.
# > \endverbatim
#
#  Arguments:
#  ==========
#
# > \param[in] SA
# > \verbatim
# >          SA is REAL
# > \endverbatim
# >
# > \param[in] SB
# > \verbatim
# >          SB is REAL
# > \endverbatim
# >
# > \param[out] C
# > \verbatim
# >          C is REAL
# > \endverbatim
# >
# > \param[out] S
# > \verbatim
# >          S is REAL
# > \endverbatim
#
#  Authors:
#  ========
#
# > \author Univ. of Tennessee
# > \author Univ. of California Berkeley
# > \author Univ. of Colorado Denver
# > \author NAG Ltd.
#
# > \date November 2017
#
# > \ingroup single_blas_level1
#
# > \par Further Details:
#  =====================
# >
# > \verbatim
# >
# >     jack dongarra, linpack, 3/11/78.
# > \endverbatim
# >
#  =====================================================================
from math import sqrt
from util import sign


def SROTG(SA, SB, C, S):
    #
    #  -- Reference BLAS level1 routine (version 3.8.0) --
    #  -- Reference BLAS is a software package provided by Univ. of Tennessee,    --
    #  -- Univ. of California Berkeley, Univ. of Colorado Denver and NAG Ltd..--
    #     November 2017
    #
    #     .. Scalar Arguments ..
    # REAL C,S,SA,SB
    #     ..
    #
    #  =====================================================================
    #
    #     .. Local Scalars ..
    # REAL R,ROE,SCALE,Z
    #     ..
    #     .. Intrinsic Functions ..
    # INTRINSIC ABS,SIGN,SQRT
    #     ..
    ROE = SB
    if abs(SA) > abs(SB):
        ROE = SA
    SCALE = abs(SA) + abs(SB)
    if SCALE == 0.0:
        C = 1.0
        S = 0.0
        R = 0.0
        Z = 0.0
    else:
        R = SCALE * sqrt((SA / SCALE) ** 2 + (SB / SCALE) ** 2)
        R = sign(ROE) * R
        C = SA / R
        S = SB / R
        Z = 1.0
        if abs(SA) > abs(SB):
            Z = S
        if abs(SB) >= abs(SA) and C != 0.0:
            Z = 1.0 / C
    SA = R
    SB = Z



# > \brief \b DCABS1
#
#  =========== DOCUMENTATION ===========
#
# Online html documentation available at
#            http://www.netlib.org/lapack/explore-html/
#
#  Definition:
#  ===========
#
#       def DCABS1(Z)
#
#       .. Scalar Arguments ..
#       COMPLEX*16 Z
#       ..
#       ..
#
#
# > \par Purpose:
#  =============
# >
# > \verbatim
# >
# > DCABS1 computes |Re(.)| + |Im(.)| of a double complex number
# > \endverbatim
#
#  Arguments:
#  ==========
#
# > \param[in] Z
# > \verbatim
# >          Z is COMPLEX*16
# > \endverbatim
#
#  Authors:
#  ========
#
# > \author Univ. of Tennessee
# > \author Univ. of California Berkeley
# > \author Univ. of Colorado Denver
# > \author NAG Ltd.
#
# > \date November 2017
#
# > \ingroup double_blas_level1
#
#  =====================================================================
def dcabs1(Z):
    #
    #  -- Reference BLAS level1 routine (version 3.8.0) --
    #  -- Reference BLAS is a software package provided by Univ. of Tennessee,    --
    #  -- Univ. of California Berkeley, Univ. of Colorado Denver and NAG Ltd..--
    #     November 2017
    #
    #     .. Scalar Arguments ..
    # COMPLEX*16 Z
    #     ..
    #     ..
    #  =====================================================================
    #
    #     .. Intrinsic Functions ..
    # INTRINSIC ABS,DBLE,DIMAG
    #
    return abs(Z.real) + abs(Z.imag)



# > \brief \b DSDOT
#
#  =========== DOCUMENTATION ===========
#
# Online html documentation available at
#            http://www.netlib.org/lapack/explore-html/
#
#  Definition:
#  ===========
#
#       def DSDOT(N,SX,INCX,SY,INCY)
#
#       .. Scalar Arguments ..
#       INTEGER INCX,INCY,N
#       ..
#       .. Array Arguments ..
#       REAL SX(*),SY(*)
#       ..
#
#    AUTHORS
#    =======
#    Lawson, C. L., (JPL), Hanson, R. J., (SNLA),
#    Kincaid, D. R., (U. of Texas), Krogh, F. T., (JPL)
#
#
# > \par Purpose:
#  =============
# >
# > \verbatim
# >
# > Compute the inner product of two vectors with extended
# > precision accumulation and result.
# >
# > Returns D.P. dot product accumulated in D.P., for S.P. SX and SY
# > DSDOT = sum for I = 0 to N-1 of  SX(LX+I*INCX) * SY(LY+I*INCY),
# > where LX = 1 if INCX >= 0, else LX = 1+(1-N)*INCX, and LY is
# > defined in a similar way using INCY.
# > \endverbatim
#
#  Arguments:
#  ==========
#
# > \param[in] N
# > \verbatim
# >          N is INTEGER
# >         number of elements in input vector(s)
# > \endverbatim
# >
# > \param[in] SX
# > \verbatim
# >          SX is REAL array, dimension(N)
# >         single precision vector with N elements
# > \endverbatim
# >
# > \param[in] INCX
# > \verbatim
# >          INCX is INTEGER
# >          storage spacing between elements of SX
# > \endverbatim
# >
# > \param[in] SY
# > \verbatim
# >          SY is REAL array, dimension(N)
# >         single precision vector with N elements
# > \endverbatim
# >
# > \param[in] INCY
# > \verbatim
# >          INCY is INTEGER
# >         storage spacing between elements of SY
# > \endverbatim
# >
# > \result DSDOT
# > \verbatim
# >          DSDOT is DOUBLE PRECISION
# >         DSDOT  double precision dot product (zero if N<=0)
# > \endverbatim
#
#  Authors:
#  ========
#
# > \author Univ. of Tennessee
# > \author Univ. of California Berkeley
# > \author Univ. of Colorado Denver
# > \author NAG Ltd.
#
# > \date December 2016
#
# > \ingroup double_blas_level1
#
# > \par Further Details:
#  =====================
# >
# > \verbatim
# > \endverbatim
#
# > \par References:
#  ================
# >
# > \verbatim
# >
# >
# >  C. L. Lawson, R. J. Hanson, D. R. Kincaid and F. T.
# >  Krogh, Basic linear algebra subprograms for Fortran
# >  usage, Algorithm No. 539, Transactions on Mathematical
# >  Software 5, 3 (September 1979), pp. 308-323.
# >
# >  REVISION HISTORY  (YYMMDD)
# >
# >  791001  DATE WRITTEN
# >  890831  Modified array declarations.  (WRB)
# >  890831  REVISION DATE from Version 3.2
# >  891214  Prologue converted to Version 4.0 format.  (BAB)
# >  920310  Corrected definition of LX in DESCRIPTION.  (WRB)
# >  920501  Reformatted the REFERENCES section.  (WRB)
# >  070118  Reformat to LAPACK style (JL)
# > \endverbatim
# >
#  =====================================================================
def DSDOT(N, SX, INCX, SY, INCY):
    #
    #  -- Reference BLAS level1 routine (version 3.7.0) --
    #  -- Reference BLAS is a software package provided by Univ. of Tennessee,    --
    #  -- Univ. of California Berkeley, Univ. of Colorado Denver and NAG Ltd..--
    #     December 2016
    #
    #     .. Scalar Arguments ..
    # INTEGER INCX,INCY,N
    #     ..
    #     .. Array Arguments ..
    # REAL SX(*),SY(*)
    #     ..
    #
    #  Authors:
    #  ========
    #  Lawson, C. L., (JPL), Hanson, R. J., (SNLA),
    #  Kincaid, D. R., (U. of Texas), Krogh, F. T., (JPL)
    #
    #  =====================================================================
    #
    #     .. Local Scalars ..
    # INTEGER I,KX,KY,NS
    #     ..
    #     .. Intrinsic Functions ..
    # INTRINSIC DBLE
    #     ..
    if N <= 0:
        return 0
    DSDOT = 0
    if INCX == INCY and INCX > 0:
        #
        #     Code for equal, positive, non-unit increments.
        #
        for I in range(0, N * INCX, INCX):
            DSDOT += (SX[I]).real * (SY(I)).real
    else:
        #
        #     Code for unequal or nonpositive increments.
        #
        KX = 1
        KY = 1
        if INCX < 0:
            KX = 1 + (1 - N) * INCX
        if INCY < 0:
            KY = 1 + (1 - N) * INCY
        for I in range(N):
            DSDOT += (SX[KX]).real * (SY[KY]).real
            KX += INCX
            KY += INCY
    return DSDOT



# > \brief \b DASUM
#
#  =========== DOCUMENTATION ===========
#
# Online html documentation available at
#            http://www.netlib.org/lapack/explore-html/
#
#  Definition:
#  ===========
#
#       def DASUM(N,DX,INCX)
#
#       .. Scalar Arguments ..
#       INTEGER INCX,N
#       ..
#       .. Array Arguments ..
#       DOUBLE PRECISION DX(*)
#       ..
#
#
# > \par Purpose:
#  =============
# >
# > \verbatim
# >
# >    DASUM takes the sum of the absolute values.
# > \endverbatim
#
#  Arguments:
#  ==========
#
# > \param[in] N
# > \verbatim
# >          N is INTEGER
# >         number of elements in input vector(s)
# > \endverbatim
# >
# > \param[in] DX
# > \verbatim
# >          DX is DOUBLE PRECISION array, dimension ( 1 + ( N - 1 )*abs( INCX ) )
# > \endverbatim
# >
# > \param[in] INCX
# > \verbatim
# >          INCX is INTEGER
# >         storage spacing between elements of DX
# > \endverbatim
#
#  Authors:
#  ========
#
# > \author Univ. of Tennessee
# > \author Univ. of California Berkeley
# > \author Univ. of Colorado Denver
# > \author NAG Ltd.
#
# > \date November 2017
#
# > \ingroup double_blas_level1
#
# > \par Further Details:
#  =====================
# >
# > \verbatim
# >
# >     jack dongarra, linpack, 3/11/78.
# >     modified 3/93 to return if incx <= 0.
# >     modified 12/3/93, array(1) declarations changed to array(*)
# > \endverbatim
# >
#  =====================================================================
def dasum(N, DX, INCX):
    #
    #  -- Reference BLAS level1 routine (version 3.8.0) --
    #  -- Reference BLAS is a software package provided by Univ. of Tennessee,    --
    #  -- Univ. of California Berkeley, Univ. of Colorado Denver and NAG Ltd..--
    #     November 2017
    #
    #     .. Scalar Arguments ..
    # INTEGER INCX,N
    #     ..
    #     .. Array Arguments ..
    # DOUBLE PRECISION DX(*)
    #     ..
    #
    #  =====================================================================
    if N <= 0 or INCX <= 0:
        return 0
    DTEMP = 0
    if INCX == 1:
        # code for increment equal to 1

        # clean-up loop
        M = N % 6
        if M != 0:
            for I in range(M):
                DTEMP += abs(DX[I])
            if N < 6:
                return DTEMP
        for I in range(M, N, 6):
            DTEMP += (
                abs(DX[I])
                + abs(DX[I + 1])
                + abs(DX[I + 2])
                + abs(DX[I + 3])
                + abs(DX[I + 4])
                + abs(DX[I + 5])
            )
    else:
        # code for increment not equal to 1
        for I in range(0, N * INCX, INCX):
            DTEMP += abs(DX[I])
    return DTEMP



# > \brief \b DROTM
#
#  =========== DOCUMENTATION ===========
#
# Online html documentation available at
#            http://www.netlib.org/lapack/explore-html/
#
#  Definition:
#  ===========
#
#       def DROTM(N,DX,INCX,DY,INCY,DPARAM)
#
#       .. Scalar Arguments ..
#       INTEGER INCX,INCY,N
#       ..
#       .. Array Arguments ..
#       DOUBLE PRECISION DPARAM(5),DX(*),DY(*)
#       ..
#
#
# > \par Purpose:
#  =============
# >
# > \verbatim
# >
# >    APPLY THE MODIFIED GIVENS TRANSFORMATION, H, TO THE 2 BY N MATRIX
# >
# >    (DX**T) , WHERE **T INDICATES TRANSPOSE. THE ELEMENTS OF DX ARE IN
# >    (DY**T)
# >
# >    DX(LX+I*INCX), I = 0 TO N-1, WHERE LX = 1 IF INCX >= 0, else:
# >    LX = (-INCX)*N, AND SIMILARLY FOR SY USING LY AND INCY.
# >    WITH DPARAM(1)=DFLAG, H HAS ONE OF THE FOLLOWING FORMS..
# >
# >    DFLAG=-1.D0     DFLAG=0.D0        DFLAG=1.D0     DFLAG=-2.D0
# >
# >      (DH11  DH12)    (1.D0  DH12)    (DH11  1.D0)    (1.D0  0.D0)
# >    H=(          )    (          )    (          )    (          )
# >      (DH21  DH22),   (DH21  1.D0),   (-1.D0 DH22),   (0.D0  1.D0).
# >    SEE DROTMG FOR A DESCRIPTION OF DATA STORAGE IN DPARAM.
# > \endverbatim
#
#  Arguments:
#  ==========
#
# > \param[in] N
# > \verbatim
# >          N is INTEGER
# >         number of elements in input vector(s)
# > \endverbatim
# >
# > \param[in,out] DX
# > \verbatim
# >          DX is DOUBLE PRECISION array, dimension ( 1 + ( N - 1 )*abs( INCX ) )
# > \endverbatim
# >
# > \param[in] INCX
# > \verbatim
# >          INCX is INTEGER
# >         storage spacing between elements of DX
# > \endverbatim
# >
# > \param[in,out] DY
# > \verbatim
# >          DY is DOUBLE PRECISION array, dimension ( 1 + ( N - 1 )*abs( INCY ) )
# > \endverbatim
# >
# > \param[in] INCY
# > \verbatim
# >          INCY is INTEGER
# >         storage spacing between elements of DY
# > \endverbatim
# >
# > \param[in] DPARAM
# > \verbatim
# >          DPARAM is DOUBLE PRECISION array, dimension (5)
# >     DPARAM(1)=DFLAG
# >     DPARAM(2)=DH11
# >     DPARAM(3)=DH21
# >     DPARAM(4)=DH12
# >     DPARAM(5)=DH22
# > \endverbatim
#
#  Authors:
#  ========
#
# > \author Univ. of Tennessee
# > \author Univ. of California Berkeley
# > \author Univ. of Colorado Denver
# > \author NAG Ltd.
#
# > \date November 2017
#
# > \ingroup double_blas_level1
#
#  =====================================================================
def DROTM(N, DX, INCX, DY, INCY, DPARAM):
    #
    #  -- Reference BLAS level1 routine (version 3.8.0) --
    #  -- Reference BLAS is a software package provided by Univ. of Tennessee,    --
    #  -- Univ. of California Berkeley, Univ. of Colorado Denver and NAG Ltd..--
    #     November 2017
    #
    #     .. Scalar Arguments ..
    # INTEGER INCX,INCY,N
    #     ..
    #     .. Array Arguments ..
    # DOUBLE PRECISION DPARAM(5),DX(*),DY(*)
    #     ..
    #
    #  =====================================================================
    #
    #     .. Local Scalars ..
    # DOUBLE PRECISION DFLAG,DH11,DH12,DH21,DH22,TWO,W,Z,ZERO
    # INTEGER I,KX,KY,NSTEPS
    #     ..
    #     .. Data statements ..
    # DATA ZERO,TWO/0.D0,2.D0/
    #     ..
    #
    DFLAG = DPARAM[1]
    if N <= 0 or (DFLAG + 2 == 0):
        return
    if INCX == INCY and INCX > 0:
        #
        NSTEPS = N * INCX
        if DFLAG < 0:
            DH11 = DPARAM[2]
            DH12 = DPARAM[4]
            DH21 = DPARAM[3]
            DH22 = DPARAM[5]
            for I in range(0, NSTEPS, INCX):
                W = DX[I]
                Z = DY[I]
                DX[I] = W * DH11 + Z * DH12
                DY[I] = W * DH21 + Z * DH22
        elif DFLAG == 0:
            DH12 = DPARAM[4]
            DH21 = DPARAM[3]
            for I in range(0, NSTEPS, INCX):
                W = DX[I]
                Z = DY[I]
                DX[I] = W + Z * DH12
                DY[I] = W * DH21 + Z
        else:
            DH11 = DPARAM[2]
            DH22 = DPARAM[5]
            for I in range(NSTEPS, INCX):
                W = DX[I]
                Z = DY[I]
                DX[I] = W * DH11 + Z
                DY[I] = -W + DH22 * Z
    else:
        KX = 1
        KY = 1
        if INCX < 0:
            KX = 1 + (1 - N) * INCX
        if INCY < 0:
            KY = 1 + (1 - N) * INCY

        if DFLAG < 0:
            DH11 = DPARAM[2]
            DH12 = DPARAM[4]
            DH21 = DPARAM[3]
            DH22 = DPARAM[5]
            for I in range(N):
                W = DX[KX]
                Z = DY[KY]
                DX[KX] = W * DH11 + Z * DH12
                DY[KY] = W * DH21 + Z * DH22
                KX += INCX
                KY += INCY
        elif DFLAG == 0:
            DH12 = DPARAM[4]
            DH21 = DPARAM[3]
            for I in range(N):
                W = DX[KX]
                Z = DY[KY]
                DX[KX] = W + Z * DH12
                DY[KY] = W * DH21 + Z
                KX += INCX
                KY += INCY
        else:
            DH11 = DPARAM[2]
            DH22 = DPARAM[5]
            for I in range(N):
                W = DX[KX]
                Z = DY[KY]
                DX[KX] = W * DH11 + Z
                DY[KY] = -W + DH22 * Z
                KX += INCX
                KY += INCY



# > \brief \b CROTG
#
#  =========== DOCUMENTATION ===========
#
# Online html documentation available at
#            http://www.netlib.org/lapack/explore-html/
#
#  Definition:
#  ===========
#
#       def CROTG(CA,CB,C,S)
#
#       .. Scalar Arguments ..
#       COMPLEX CA,CB,S
#       REAL C
#       ..
#
#
# > \par Purpose:
#  =============
# >
# > \verbatim
# >
# > CROTG determines a complex Givens rotation.
# > \endverbatim
#
#  Arguments:
#  ==========
#
# > \param[in] CA
# > \verbatim
# >          CA is COMPLEX
# > \endverbatim
# >
# > \param[in] CB
# > \verbatim
# >          CB is COMPLEX
# > \endverbatim
# >
# > \param[out] C
# > \verbatim
# >          C is REAL
# > \endverbatim
# >
# > \param[out] S
# > \verbatim
# >          S is COMPLEX
# > \endverbatim
#
#  Authors:
#  ========
#
# > \author Univ. of Tennessee
# > \author Univ. of California Berkeley
# > \author Univ. of Colorado Denver
# > \author NAG Ltd.
#
# > \date November 2017
#
# > \ingroup complex_blas_level1
#
#  =====================================================================
from math import sqrt


def CROTG(CA, CB, C, S):
    #
    #  -- Reference BLAS level1 routine (version 3.8.0) --
    #  -- Reference BLAS is a software package provided by Univ. of Tennessee,    --
    #  -- Univ. of California Berkeley, Univ. of Colorado Denver and NAG Ltd..--
    #     November 2017
    #
    #     .. Scalar Arguments ..
    # COMPLEX CA,CB,S
    # REAL C
    #     ..
    #
    #  =====================================================================
    #
    #     .. Local Scalars ..
    # COMPLEX ALPHA
    # REAL NORM,SCALE
    #     ..
    #     .. Intrinsic Functions ..
    # INTRINSIC CABS,CONJG,SQRT
    #     ..
    if abs(CA) == 0.0:
        C = 0.0
        S = 1 + 0j
        CA = CB
    else:
        SCALE = abs(CA) + abs(CB)
        NORM = SCALE * sqrt((abs(CA / SCALE)) ** 2 + (abs(CB / SCALE)) ** 2)
        ALPHA = CA / abs(CA)
        C = abs(CA) / NORM
        S = ALPHA * (CB).conjugate() / NORM
        CA = ALPHA * NORM



# > \brief \b SROTMG
#
#  =========== DOCUMENTATION ===========
#
# Online html documentation available at
#            http://www.netlib.org/lapack/explore-html/
#
#  Definition:
#  ===========
#
#       def SROTMG(SD1,SD2,SX1,SY1,SPARAM)
#
#       .. Scalar Arguments ..
#       REAL SD1,SD2,SX1,SY1
#       ..
#       .. Array Arguments ..
#       REAL SPARAM(5)
#       ..
#
#
# > \par Purpose:
#  =============
# >
# > \verbatim
# >
# >    CONSTRUCT THE MODIFIED GIVENS TRANSFORMATION MATRIX H WHICH ZEROS
# >    THE SECOND COMPONENT OF THE 2-VECTOR  (SQRT(SD1)*SX1,SQRT(SD2)*>    SY2)**T.
# >    WITH SPARAM(1)=SFLAG, H HAS ONE OF THE FOLLOWING FORMS..
# >
# >    SFLAG=-1.E0     SFLAG=0.E0        SFLAG=1.E0     SFLAG=-2.E0
# >
# >      (SH11  SH12)    (1.E0  SH12)    (SH11  1.E0)    (1.E0  0.E0)
# >    H=(          )    (          )    (          )    (          )
# >      (SH21  SH22),   (SH21  1.E0),   (-1.E0 SH22),   (0.E0  1.E0).
# >    LOCATIONS 2-4 OF SPARAM CONTAIN SH11,SH21,SH12, AND SH22
# >    RESPECTIVELY. (VALUES OF 1.E0, -1.E0, OR 0.E0 IMPLIED BY THE
# >    VALUE OF SPARAM(1) ARE NOT STORED IN SPARAM.)
# >
# >    THE VALUES OF GAMSQ AND RGAMSQ SET IN THE DATA STATEMENT MAY BE
# >    INEXACT.  THIS IS OK AS THEY ARE ONLY USED FOR TESTING THE SIZE
# >    OF SD1 AND SD2.  ALL ACTUAL SCALING OF DATA IS DONE USING GAM.
# >
# > \endverbatim
#
#  Arguments:
#  ==========
#
# > \param[in,out] SD1
# > \verbatim
# >          SD1 is REAL
# > \endverbatim
# >
# > \param[in,out] SD2
# > \verbatim
# >          SD2 is REAL
# > \endverbatim
# >
# > \param[in,out] SX1
# > \verbatim
# >          SX1 is REAL
# > \endverbatim
# >
# > \param[in] SY1
# > \verbatim
# >          SY1 is REAL
# > \endverbatim
# >
# > \param[out] SPARAM
# > \verbatim
# >          SPARAM is REAL array, dimension (5)
# >     SPARAM(1)=SFLAG
# >     SPARAM(2)=SH11
# >     SPARAM(3)=SH21
# >     SPARAM(4)=SH12
# >     SPARAM(5)=SH22
# > \endverbatim
#
#  Authors:
#  ========
#
# > \author Univ. of Tennessee
# > \author Univ. of California Berkeley
# > \author Univ. of Colorado Denver
# > \author NAG Ltd.
#
# > \date November 2017
#
# > \ingroup single_blas_level1
#
#  =====================================================================
def SROTMG(SD1, SD2, SX1, SY1, SPARAM):
    #
    #  -- Reference BLAS level1 routine (version 3.8.0) --
    #  -- Reference BLAS is a software package provided by Univ. of Tennessee,    --
    #  -- Univ. of California Berkeley, Univ. of Colorado Denver and NAG Ltd..--
    #     November 2017
    #
    #     .. Scalar Arguments ..
    # REAL SD1,SD2,SX1,SY1
    #     ..
    #     .. Array Arguments ..
    # REAL SPARAM(5)
    #     ..
    #
    #  =====================================================================
    #
    #     .. Local Scalars ..
    #    REAL GAM,GAMSQ,ONE,RGAMSQ,SFLAG,SH11,SH12,SH21,SH22,SP1,SP2,SQ1,
    #   $     SQ2,STEMP,SU,TWO,ZERO
    #     ..
    #     .. Intrinsic Functions ..
    # INTRINSIC ABS
    #     ..
    #     .. Data statements ..
    #
    # DATA ZERO,ONE,TWO/0.E0,1.E0,2.E0/
    # DATA GAM,GAMSQ,RGAMSQ/4096.E0,1.67772E7,5.96046E-8/
    #     ..
    GAM = 4096
    GAMSQ = 1.67772e7
    RGAMSQ = 5.96046e-8
    if SD1 < 0:
        #        GO ZERO-H-D-AND-SX1..
        SFLAG = -1
        SH11 = 0
        SH12 = 0
        SH21 = 0
        SH22 = 0
        #
        SD1 = 0
        SD2 = 0
        SX1 = 0
    else:
        #        CASE-SD1-NONNEGATIVE
        SP2 = SD2 * SY1
        if SP2 == 0:
            SFLAG = -2
            SPARAM[1] = SFLAG
            return
        #        REGULAR-CASE..
        SP1 = SD1 * SX1
        SQ2 = SP2 * SY1
        SQ1 = SP1 * SX1
        #
        if abs(SQ1) > abs(SQ2):
            SH21 = -SY1 / SX1
            SH12 = SP2 / SP1
            #
            SU = 1 - SH12 * SH21
            #
            if SU > 0:
                SFLAG = 0
                SD1 = SD1 / SU
                SD2 = SD2 / SU
                SX1 = SX1 * SU
        else:

            if SQ2 < 0:
                #              GO ZERO-H-D-AND-SX1..
                SFLAG = -1
                SH11 = 0
                SH12 = 0
                SH21 = 0
                SH22 = 0
                #
                SD1 = 0
                SD2 = 0
                SX1 = 0
            else:
                SFLAG = 1
                SH11 = SP1 / SP2
                SH22 = SX1 / SY1
                SU = 1 + SH11 * SH22
                STEMP = SD2 / SU
                SD2 = SD1 / SU
                SD1 = STEMP
                SX1 = SY1 * SU

        #     PROCESURE..SCALE-CHECK
        if SD1 != 0:
            while (SD1 <= RGAMSQ) or (SD1 >= GAMSQ):
                if SFLAG == 0:
                    SH11 = 1
                    SH22 = 1
                    SFLAG = -1
                else:
                    SH21 = -1
                    SH12 = 1
                    SFLAG = -1
                if SD1 <= RGAMSQ:
                    SD1 = SD1 * GAM ** 2
                    SX1 = SX1 / GAM
                    SH11 = SH11 / GAM
                    SH12 = SH12 / GAM
                else:
                    SD1 = SD1 / GAM ** 2
                    SX1 = SX1 * GAM
                    SH11 = SH11 * GAM
                    SH12 = SH12 * GAM

        if SD2 != 0:
            while (abs(SD2) <= RGAMSQ) or (abs(SD2) >= GAMSQ):
                if SFLAG == 0:
                    SH11 = 1
                    SH22 = 1
                    SFLAG = -1
                else:
                    SH21 = -1
                    SH12 = 1
                    SFLAG = -1
                if abs(SD2) <= RGAMSQ:
                    SD2 = SD2 * GAM ** 2
                    SH21 = SH21 / GAM
                    SH22 = SH22 / GAM
                else:
                    SD2 = SD2 / GAM ** 2
                    SH21 = SH21 * GAM
                    SH22 = SH22 * GAM

    if SFLAG < 0:
        SPARAM[2] = SH11
        SPARAM[3] = SH21
        SPARAM[4] = SH12
        SPARAM[5] = SH22
    elif SFLAG == 0:
        SPARAM[3] = SH21
        SPARAM[4] = SH12
    else:
        SPARAM[2] = SH11
        SPARAM[5] = SH22

    SPARAM[1] = SFLAG



# > \brief \b DCOPY
#
#  =========== DOCUMENTATION ===========
#
# Online html documentation available at
#            http://www.netlib.org/lapack/explore-html/
#
#  Definition:
#  ===========
#
#       def DCOPY(N,DX,INCX,DY,INCY)
#
#       .. Scalar Arguments ..
#       INTEGER INCX,INCY,N
#       ..
#       .. Array Arguments ..
#       DOUBLE PRECISION DX(*),DY(*)
#       ..
#
#
# > \par Purpose:
#  =============
# >
# > \verbatim
# >
# >    DCOPY copies a vector, x, to a vector, y.
# >    uses unrolled loops for increments equal to 1.
# > \endverbatim
#
#  Arguments:
#  ==========
#
# > \param[in] N
# > \verbatim
# >          N is INTEGER
# >         number of elements in input vector(s)
# > \endverbatim
# >
# > \param[in] DX
# > \verbatim
# >          DX is DOUBLE PRECISION array, dimension ( 1 + ( N - 1 )*abs( INCX ) )
# > \endverbatim
# >
# > \param[in] INCX
# > \verbatim
# >          INCX is INTEGER
# >         storage spacing between elements of DX
# > \endverbatim
# >
# > \param[out] DY
# > \verbatim
# >          DY is DOUBLE PRECISION array, dimension ( 1 + ( N - 1 )*abs( INCY ) )
# > \endverbatim
# >
# > \param[in] INCY
# > \verbatim
# >          INCY is INTEGER
# >         storage spacing between elements of DY
# > \endverbatim
#
#  Authors:
#  ========
#
# > \author Univ. of Tennessee
# > \author Univ. of California Berkeley
# > \author Univ. of Colorado Denver
# > \author NAG Ltd.
#
# > \date November 2017
#
# > \ingroup double_blas_level1
#
# > \par Further Details:
#  =====================
# >
# > \verbatim
# >
# >     jack dongarra, linpack, 3/11/78.
# >     modified 12/3/93, array(1) declarations changed to array(*)
# > \endverbatim
# >
#  =====================================================================
from ..util import slice_


def dcopy(N, DX, INCX, DY, INCY):
    """
    FIXME
    """
    #
    #  -- Reference BLAS level1 routine (version 3.8.0) --
    #  -- Reference BLAS is a software package provided by Univ. of Tennessee,    --
    #  -- Univ. of California Berkeley, Univ. of Colorado Denver and NAG Ltd..--
    #     November 2017
    #
    #     .. Scalar Arguments ..
    # INTEGER INCX,INCY,N
    #     ..
    #     .. Array Arguments ..
    # DOUBLE PRECISION DX(*),DY(*)
    #     ..
    #
    #  =====================================================================

    if N <= 0:
        return
    DY[slice_(N, INCY)] = DX[slice_(N, INCX)]



# > \brief \b CSROT
#
#  =========== DOCUMENTATION ===========
#
# Online html documentation available at
#            http://www.netlib.org/lapack/explore-html/
#
#  Definition:
#  ===========
#
#       def CSROT( N, CX, INCX, CY, INCY, C, S )
#
#       .. Scalar Arguments ..
#       INTEGER           INCX, INCY, N
#       REAL              C, S
#       ..
#       .. Array Arguments ..
#       COMPLEX           CX( * ), CY( * )
#       ..
#
#
# > \par Purpose:
#  =============
# >
# > \verbatim
# >
# > CSROT applies a plane rotation, where the cos and sin (c and s) are real
# > and the vectors cx and cy are complex.
# > jack dongarra, linpack, 3/11/78.
# > \endverbatim
#
#  Arguments:
#  ==========
#
# > \param[in] N
# > \verbatim
# >          N is INTEGER
# >           On entry, N specifies the order of the vectors cx and cy.
# >           N must be at least zero.
# > \endverbatim
# >
# > \param[in,out] CX
# > \verbatim
# >          CX is COMPLEX array, dimension at least
# >           ( 1 + ( N - 1 )*abs( INCX ) ).
# >           Before entry, the incremented array CX must contain the n
# >           element vector cx. On exit, CX is overwritten by the updated
# >           vector cx.
# > \endverbatim
# >
# > \param[in] INCX
# > \verbatim
# >          INCX is INTEGER
# >           On entry, INCX specifies the increment for the elements of
# >           CX. INCX must not be zero.
# > \endverbatim
# >
# > \param[in,out] CY
# > \verbatim
# >          CY is COMPLEX array, dimension at least
# >           ( 1 + ( N - 1 )*abs( INCY ) ).
# >           Before entry, the incremented array CY must contain the n
# >           element vector cy. On exit, CY is overwritten by the updated
# >           vector cy.
# > \endverbatim
# >
# > \param[in] INCY
# > \verbatim
# >          INCY is INTEGER
# >           On entry, INCY specifies the increment for the elements of
# >           CY. INCY must not be zero.
# > \endverbatim
# >
# > \param[in] C
# > \verbatim
# >          C is REAL
# >           On entry, C specifies the cosine, cos.
# > \endverbatim
# >
# > \param[in] S
# > \verbatim
# >          S is REAL
# >           On entry, S specifies the sine, sin.
# > \endverbatim
#
#  Authors:
#  ========
#
# > \author Univ. of Tennessee
# > \author Univ. of California Berkeley
# > \author Univ. of Colorado Denver
# > \author NAG Ltd.
#
# > \date December 2016
#
# > \ingroup complex_blas_level1
#
#  =====================================================================
def csrot(N, CX, INCX, CY, INCY, C, S):
    #
    #  -- Reference BLAS level1 routine (version 3.7.0) --
    #  -- Reference BLAS is a software package provided by Univ. of Tennessee,    --
    #  -- Univ. of California Berkeley, Univ. of Colorado Denver and NAG Ltd..--
    #     December 2016
    #
    #     .. Scalar Arguments ..
    # INTEGER           INCX, INCY, N
    # REAL              C, S
    #     ..
    #     .. Array Arguments ..
    # COMPLEX           CX( * ), CY( * )
    #     ..
    #
    #  =====================================================================
    if N <= 0:
        return
    if INCX == 1 and INCY == 1:
        # code for both increments equal to 1
        for I in range(N):
            CTEMP = C * CX[I] + S * CY[I]
            CY[I] = C * CY[I] - S * CX[I]
            CX[I] = CTEMP
    else:
        # code for unequal increments or equal increments not equal to 1
        IX = 1
        IY = 1
        if INCX < 0:
            IX = (-N + 1) * INCX + 1
        if INCY < 0:
            IY = (-N + 1) * INCY + 1
        for I in range(N):
            CTEMP = C * CX[IX] + S * CY[IY]
            CY[IY] = C * CY[IY] - S * CX[IX]
            CX[IX] = CTEMP
            IX += INCX
            IY += INCY
